// LGhost.dll v0.3.01 Copyright(C) 2002, 2003 minamina
// Avisynth Plugin - Ghost Reduction (YUY2 and YV12 Only, Luminance Ghost Only)

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef _LUMINANCE_GHOST_H
#define _LUMINANCE_GHOST_H

#ifdef _AVISYNTH_25
	#define _AVISYNTH_H "avisynth.h"
	#define _AVISYNTH_PLUGIN_INIT AvisynthPluginInit2
#else // _AVISYNTH_25
	#define _AVISYNTH_H "avisynth1.h"
	#define _AVISYNTH_PLUGIN_INIT AvisynthPluginInit
#endif // _AVISYNTH_25

#define _USE_MAKE_WRITABLE

#ifdef _MSC_VER
	#include <crtdbg.h>
#else
	#define _ASSERTE(x) assert(x)
	#include <assert.h>
#endif

#include <windows.h>
#include <memory.h>
#include <math.h>
#include <assert.h>
#pragma warning (push, 3)
#include <boost/shared_ptr.hpp>
#include <boost/lexical_cast.hpp>
#include _AVISYNTH_H
#pragma warning (pop)

template<class T> T ROUND(const T& n, const T& low, const T& high)
{
	return min(max(n, low), high);
}

#define ROUND_BYTE(n, low, high) static_cast<BYTE>(ROUND(n, low, high))

class LuminanceGhost : public GenericVideoFilter
{
private:
	enum
	{
		INTENSITY_SHIFT = 7,
		MAX_INTENSITY = (1 << INTENSITY_SHIFT),
		MAX_PROCESS = 18,
	};

	enum
	{
		MODE_NOT_USE = 0,
		MODE_EDGE,
		MODE_LUMINANCE,
		MODE_EDGE_RISING,
		MODE_EDGE_FALLING,
	};

	typedef struct
	{
		int nShift;
		int nIntensity;
		int nStartX;
		int nEndX;
	} OPTION_DATA;

	int m_nMaxEdgeFalling;
	int m_nMaxEdgeRising;
	int m_nMaxLuminance;
	int m_nMaxEdge;

	OPTION_DATA m_optionEdgeFalling[MAX_PROCESS];
	OPTION_DATA m_optionEdgeRising[MAX_PROCESS];
	OPTION_DATA m_optionLuminance[MAX_PROCESS];
	OPTION_DATA m_optionEdge[MAX_PROCESS];

public:
	static void PluginInit(IScriptEnvironment *env)
	{ env->AddFunction(GetName(), GetParams(), Create, 0); }

	LuminanceGhost(PClip _child, AVSValue args, IScriptEnvironment* env)
		: GenericVideoFilter(_child)
	{
		if(vi.IsYUY2()) {}
#ifdef _AVISYNTH_25
		else if (vi.IsYV12()) {}
#endif // _AVISYNTH_25
		else
		{
			env->ThrowError("%s : YUY2, YV12 only", GetName());
		}

		int nStep = vi.IsYUY2() + 1;

		int nArraySize = args.ArraySize() / 3;
		if (nArraySize > MAX_PROCESS)
		{
			env->ThrowError("%s : too many options. (Max %d)", GetName(), MAX_PROCESS);
		}

		m_nMaxEdgeFalling = 0;
		m_nMaxEdgeRising = 0;
		m_nMaxLuminance = 0;
		m_nMaxEdge = 0;
		int i;
		for (i = 0; i < nArraySize; i++)
		{
			AVSValue arg(&args[i * 3], 3);
			int nMode = arg[0].AsInt(MODE_NOT_USE);
			int nShift = arg[1].AsInt(0);
			int nIntensity = arg[2].AsInt(0);
			int nWidth = vi.width;

			if (nWidth < abs(nShift))
			{
				env->ThrowError("%s : -%d < shift < %d (clip width)", GetName(), nWidth, nWidth);
			}
			if (-MAX_INTENSITY > nIntensity || nIntensity >= MAX_INTENSITY)
			{
				env->ThrowError("%s : -%d < intensity < %d", GetName(), MAX_INTENSITY, MAX_INTENSITY - 1);
			}

			if (nIntensity != 0)
			{
				switch (nMode)
				{
				case MODE_LUMINANCE:
					{
						m_optionLuminance[m_nMaxLuminance].nShift = nShift * nStep;
						m_optionLuminance[m_nMaxLuminance].nIntensity = nIntensity;
						m_optionLuminance[m_nMaxLuminance].nStartX = max(0, nShift) * nStep;
						m_optionLuminance[m_nMaxLuminance].nEndX = min(nWidth, nWidth + nShift) * nStep;
						m_nMaxLuminance++;
					}
					break;
				case MODE_EDGE:
					{
						if (nShift != 0)
						{
							if (nShift < 0)
							{
								nShift++;
								nWidth--;
//								nIntensity = -nIntensity;
							}
							m_optionEdge[m_nMaxEdge].nShift = nShift * nStep;
							m_optionEdge[m_nMaxEdge].nIntensity = nIntensity;
							m_optionEdge[m_nMaxEdge].nStartX = max(0, nShift) * nStep;
							m_optionEdge[m_nMaxEdge].nEndX = min(nWidth, nWidth + nShift) * nStep;
							m_nMaxEdge++;
						}
					}
					break;
				case MODE_EDGE_RISING:
					{
						if (nShift != 0)
						{
							if (nShift < 0)
							{
								nShift++;
								nWidth--;
//								nIntensity = -nIntensity;
							}
							m_optionEdgeRising[m_nMaxEdgeRising].nShift = nShift * nStep;
							m_optionEdgeRising[m_nMaxEdgeRising].nIntensity = nIntensity;
							m_optionEdgeRising[m_nMaxEdgeRising].nStartX = max(0, nShift) * nStep;
							m_optionEdgeRising[m_nMaxEdgeRising].nEndX = min(nWidth, nWidth + nShift) * nStep;
							m_nMaxEdgeRising++;
						}
					}
					break;
				case MODE_EDGE_FALLING:
					{
						if (nShift != 0)
						{
							if (nShift < 0)
							{
								nShift++;
								nWidth--;
//								nIntensity = -nIntensity;
							}
							m_optionEdgeFalling[m_nMaxEdgeFalling].nShift = nShift * nStep;
							m_optionEdgeFalling[m_nMaxEdgeFalling].nIntensity = nIntensity;
							m_optionEdgeFalling[m_nMaxEdgeFalling].nStartX = max(0, nShift) * nStep;
							m_optionEdgeFalling[m_nMaxEdgeFalling].nEndX = min(nWidth, nWidth + nShift) * nStep;
							m_nMaxEdgeFalling++;
						}
					}
					break;
				case MODE_NOT_USE:
				default:
//					__assume(0);
					break;
				}
			}
		}
	}

private:
	static AVSValue __cdecl Create(AVSValue args, void* /*user_data*/, IScriptEnvironment *env)
	{ return new LuminanceGhost(args[0].AsClip(), args[1], env); }
	static const char *GetName() { return "LGhost"; }
	static const char *GetParams() { return "c.*"; }

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment *env)
	{
		PVideoFrame frameSrc(child->GetFrame(n, env));
#ifdef _USE_MAKE_WRITABLE
		PVideoFrame frameDest = frameSrc;
		env->MakeWritable(&frameDest);
#else // _USE_MAKE_WRITABLE
		PVideoFrame frameDest(env->NewVideoFrame(vi));
#endif // _USE_MAKE_WRITABLE

		const BYTE* pSrc = frameSrc->GetReadPtr();
		BYTE* pDest = frameDest->GetWritePtr();

//		const int nWidth = vi.width;
		const int nHeight = vi.height;
		const int nPitch = frameSrc->GetPitch();
		const int nRowSize = frameSrc->GetRowSize();

#ifdef _USE_MAKE_WRITABLE
#else // _USE_MAKE_WRITABLE
			env->BitBlt(pDest, nPitch, pSrc, nPitch, nRowSize, nHeight);
	#ifdef _AVISYNTH_25
			env->BitBlt(frameDest->GetWritePtr(PLANAR_V), frameDest->GetPitch(PLANAR_V), frameSrc->GetReadPtr(PLANAR_V), frameSrc->GetPitch(PLANAR_V), frameSrc->GetRowSize(PLANAR_V), frameSrc->GetHeight(PLANAR_V));
			env->BitBlt(frameDest->GetWritePtr(PLANAR_U), frameDest->GetPitch(PLANAR_U), frameSrc->GetReadPtr(PLANAR_U), frameSrc->GetPitch(PLANAR_U), frameSrc->GetRowSize(PLANAR_U), frameSrc->GetHeight(PLANAR_U));
	#endif // _AVISYNTH_25
#endif // _USE_MAKE_WRITABLE

		int nStep = vi.IsYUY2() + 1;

		int *pBuffer = reinterpret_cast<int*>(::malloc(sizeof(int) * nRowSize));
		if (pBuffer != NULL)
		{
			int nX, nY, i;
			for (nY = nHeight; nY; nY--)
			{
				::ZeroMemory(pBuffer, sizeof(int) * nRowSize);
				int nShift, nIntensity, nStartX, nEndX, nTempEdge;
				for (i = 0; i < m_nMaxLuminance; i++)
				{
					nShift = m_optionLuminance[i].nShift;
					nIntensity = m_optionLuminance[i].nIntensity;
					nStartX = m_optionLuminance[i].nStartX;
					nEndX = m_optionLuminance[i].nEndX;
					for (nX = nStartX; nX < nEndX; nX += nStep)
					{
						pBuffer[nX] += pSrc[nX - nShift] * nIntensity;
					}
				}
				for (i = 0; i < m_nMaxEdge; i++)
				{
					nShift = m_optionEdge[i].nShift;
					nIntensity = m_optionEdge[i].nIntensity;
					nStartX = m_optionEdge[i].nStartX;
					nEndX = m_optionEdge[i].nEndX;
					for (nX = nStartX; nX < nEndX; nX += nStep)
					{
						pBuffer[nX] += (pSrc[nX - nShift + nStep] - pSrc[nX - nShift]) * nIntensity;
					}
				}
				for (i = 0; i < m_nMaxEdgeRising; i++)
				{
					nShift = m_optionEdgeRising[i].nShift;
					nIntensity = m_optionEdgeRising[i].nIntensity;
					nStartX = m_optionEdgeRising[i].nStartX;
					nEndX = m_optionEdgeRising[i].nEndX;
					for (nX = nStartX; nX < nEndX; nX += nStep)
					{
						nTempEdge = pSrc[nX - nShift + nStep] - pSrc[nX - nShift];
						if (nTempEdge > 0)
						{
							pBuffer[nX] += nTempEdge * nIntensity;
						}
					}
				}
				for (i = 0; i < m_nMaxEdgeFalling; i++)
				{
					nShift = m_optionEdgeFalling[i].nShift;
					nIntensity = m_optionEdgeFalling[i].nIntensity;
					nStartX = m_optionEdgeFalling[i].nStartX;
					nEndX = m_optionEdgeFalling[i].nEndX;
					for (nX = nStartX; nX < nEndX; nX += nStep)
					{
						nTempEdge = pSrc[nX - nShift + nStep] - pSrc[nX - nShift];
						if (nTempEdge < 0)
						{
							pBuffer[nX] += nTempEdge * nIntensity;
						}
					}
				}
				for (nX = 0; nX < nRowSize; nX += nStep)
				{
					pDest[nX] = ROUND_BYTE(pSrc[nX] + (pBuffer[nX] >> INTENSITY_SHIFT), 0, 255);
				}
				pSrc += nPitch;
				pDest += nPitch;
			}
			::free(pBuffer);
		}

		return frameDest;
	}
};

#endif // _LUMINANCE_GHOST_H
